from functools import update_wrapper


def click_options(settings_cls, appname, config_files):
    try:
        import click
    except ImportError as e:
        raise ModuleNotFoundError(
            'You need to install "click" to use this feature'
        ) from e

    def pass_settings(f):
        """Similar to :func:`pass_context`, but only pass the object on the
        context onwards (:attr:`Context.obj`).  This is useful if that object
        represents the state of a nested system.
        """

        def new_func(*args, **kwargs):
            return f(
                click.get_current_context().obj["settings"], *args, **kwargs
            )

        return update_wrapper(new_func, f)

    def cb(c, p, v):
        if c.obj is None:
            c.obj = {}
        if "settings" not in c.obj:
            c.obj[
                "settings"
            ] = {}  # typed_settings.load_settings(cls=settings_cls)
        c.obj["settings"][p.name] = v
        return v

    def wrap(f):
        f = click.option(
            "--spam", callback=cb, expose_value=False, is_eager=True
        )(f)
        f = click.option(
            "--eggs", callback=cb, expose_value=False, is_eager=True
        )(f)
        f = pass_settings(f)
        return f

    return wrap



"""
Core functionality for loading settings.
"""
import os
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
)

import attr
import toml

from ._dict_utils import (
    FieldList,
    _deep_fields,
    _get_path,
    _merge_dicts,
    _set_path,
)


class _Auto:
    """
    Sentinel class to indicate the lack of a value when ``None`` is ambiguous.

    ``_Auto`` is a singleton. There is only ever one of it.
    """

    _singleton = None

    def __new__(cls):
        if _Auto._singleton is None:
            _Auto._singleton = super(_Auto, cls).__new__(cls)
        return _Auto._singleton

    def __repr__(self):
        return "AUTO"


AUTO = _Auto()
"""
Sentinel to indicate the lack of a value when ``None`` is ambiguous.
"""

T = TypeVar("T")


def load_settings(
    settings_cls: Type[T],
    appname: str,
    config_files: Iterable[Union[str, Path]] = (),
    *,
    config_file_section: Union[str, _Auto] = AUTO,
    config_files_var: Union[None, str, _Auto] = AUTO,
    env_prefix: Union[None, str, _Auto] = AUTO,
) -> T:
    """
    Loads settings for *appname* and returns an instance of *settings_cls*

    Settings can be loaded from *config_files* and/or from the files specified
    via the *config_files_var* environment variable.  Settings can also be
    overridden via environment variables named like the corresponding setting
    and prefixed with *env_prefix*.

    Settings precedence (from lowest to highest priority):

    - Default value from *settings_cls*
    - First file from *config_files*
    - ...
    - Last file from *config_files*
    - First file from *config_files_var*
    - ...
    - Last file from *config_files_var*
    - Environment variable *env_prefix*_{SETTING}

    Config files (both, explicitly specified, and loaded from an environment
    variable) are optional by default.  You can prepend an ``!`` to their path
    to mark them as mandatory (e.g., `!/etc/credentials.toml`).  An error is
    raised if a mandatory file does not exist.

    Args:
        settings_cls: Attrs class with default settings.

        appname: Your application's name.  Used to derive defaults for the
          remaining args.

        config_files: Load settings from these TOML files.

        config_file_section: Name of your app's section in the config file.
          By default, use *appname*.

        config_files_var: Load list of settings files from this environment
          variable.  By default, use *APPNAME*_SETTINGS.  Multiple paths have
          to be separated by ":".  Each settings file will update its
          predecessor, so the last file will have the highest priority.

          Set to ``None`` to disable this feature.

        env_prefix: Load settings from environment variables with this prefix.
          By default, use *APPNAME_*.  Set to ``None`` to disable loading env
          vars.

    Returns:
        An instance of *settings_cls* populated with settings from settings
        files and environment variables.

    Raises:
        FileNotFoundError: If a mandatory config file does not exist.
        TypeError: If config values cannot be converted to the required type.
        ValueError: If config values don't meet their requirements.
    """
    fields = _deep_fields(settings_cls)
    settings = _load_settings(
        fields=fields,
        appname=appname,
        config_files=config_files,
        config_file_section=config_file_section,
        config_files_var=config_files_var,
        env_prefix=env_prefix,
    )
    return settings_cls(**settings)  # type: ignore


def update_settings(settings: T, path: str, value: Any) -> T:
    """Returns a copy of *settings* with an updated *value* at *path*.

    Args:
        settings: An instance of a settings class.
        path: A dot-separated path to the setting to update.
        value: The new value to set to the attribute at *path*.

    Returns:
        A copy of *settings* with the updated value.

    Raises:
        AttributeError: *path* does not point to an existing attribute.
    """
    current = settings
    for name in path.split("."):
        try:
            print(current, name)
            current = getattr(current, name)
        except AttributeError:
            raise AttributeError(
                f"'{type(settings).__name__}' object has no setting '{path}'"
            ) from None

    settings_dict = attr.asdict(settings)
    _set_path(settings_dict, path, value)
    return type(settings)(**settings_dict)  # type: ignore


def _load_settings(
    *,
    fields: FieldList,
    appname: str,
    config_files: Iterable[Union[str, Path]],
    config_file_section: Union[str, _Auto],
    config_files_var: Union[None, str, _Auto],
    env_prefix: Union[None, str, _Auto],
) -> Dict[str, Any]:
    """
    Loads settings for *fields* and returns them as dict.

    This function makes it easier to extend settings since it returns a dict
    that can easily be updated (as compared to of frozen settings instances).

    See :func:`load_settings() for details on the arguments.
    """
    loaded_settings = [
        _from_toml(
            fields=fields,
            appname=appname,
            files=config_files,
            section=config_file_section,
            var_name=config_files_var,
        ),
        # _from_dotenv(),
        _from_env(fields=fields, appname=appname, prefix=env_prefix),
    ]
    settings: Dict[str, Any] = {}
    for ls in loaded_settings:
        _merge_dicts(settings, ls)
    return settings


def _from_toml(
    fields: FieldList,
    appname: str,
    files: Iterable[Union[str, Path]],
    section: Union[str, _Auto],
    var_name: Union[None, str, _Auto],
) -> Dict[str, Any]:
    """
    Loads settings from toml files.

    Settings of multiple files will be merged.  The last file has the highest
    precedence.

    Args:
        fields: The list of available settings.
        appname: Appname to derive *section* and *var_name* from.
        files: A list of filenames to try to load.
        section: The name of the TOML file section to load data from.  Will be
          :code:`{appname}` if it is :data:`AUTO`.
        var_name: Name of the environment variable that may hold additional
          file paths.  Will be :code:`{APPNAME}_SETTINGS` if it is
          :data:`AUTO`.

    Returns:
        A dict with the loaded settings.
    """
    section = appname if isinstance(section, _Auto) else section
    var_name = (
        f"{appname.upper()}_SETTINGS"
        if isinstance(var_name, _Auto)
        else var_name
    )

    paths = _get_config_filenames(files, var_name)
    settings: Dict[str, Any] = {}
    for path in paths:
        toml_settings = _load_toml(fields, path, section)
        _merge_dicts(settings, toml_settings)
    return settings


def _get_config_filenames(
    config_files: Iterable[Union[str, Path]],
    config_files_var: Optional[str],
) -> List[Path]:
    """
    Concatenates *config_files* and files from env var *config_files_var*.

    Mandatory files can be prefixed with ``!``.  Optional files will be
    stripped from the result if they don't exist.

    Args:
        config_files: A list of paths to settings files.
        config_files_var: The name of the environment variable that may held
          additional settings file names.

    Returns:
        A list of paths to existing config files.  Paths of files that don't
        exist are stripped from the input.
    """
    candidates = [str(f) for f in config_files]
    if config_files_var:
        candidates += os.getenv(config_files_var, "").split(":")

    paths = []
    for f in candidates:
        if f and f[0] == "!":
            # Always add mandatory files
            paths.append(Path(f[1:]))
        else:
            # Add optional files only if they exist
            p = Path(f)
            if p.is_file():
                paths.append(p)

    return paths


def _load_toml(fields: FieldList, path: Path, section: str) -> Dict[str, Any]:
    """
    Loads settings from a TOML file and returns them.

    Args:
        fields: The list of available settings.
        path: The path to the config file.
        section: The config file section to load settings from.

    Returns:
        A dict with the loaded settings.

    Raises:
        FileNotFoundError: If *path* does not exist.
    """
    sections = section.split(".")
    settings = toml.load(path.open())
    for s in sections:
        try:
            settings = settings[s]
        except KeyError:
            return {}
    settings = _rename_dict_keys(settings)
    settings = _clean_settings(fields, settings)
    return settings


def _rename_dict_keys(d: Mapping[str, Any]) -> Dict[str, Any]:
    """
    Recursively replaces "-" in dict keys with "_".

    Args:
        d: The input dict.

    Returns:
        A newly created dict with the renamed keys.
    """
    result = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = _rename_dict_keys(v)
        result[k.replace("-", "_")] = v
    return result


def _clean_settings(
    fields: FieldList, settings: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Recursively remove invalid entries from *settings* and return a new dict.

    Args:
        fields: The list of available settings.
        settings: The settings to be cleaned.

    Returns:
        A newly created dict with the cleaned settings.
    """
    cleaned: Dict[str, Any] = {}
    for path, _field, _cls in fields:
        try:
            val = _get_path(settings, path)
        except KeyError:
            continue
        _set_path(cleaned, path, val)
    return cleaned


def _from_env(
    fields: FieldList, appname: str, prefix: Union[None, str, _Auto]
) -> Dict[str, Any]:
    """
    Loads settings from environment variables.

    Args:
        fields: The list of available settings.
        appname: Appname to derive *prefix* from.
        prefix: Prefix for environment variables.  Will be :code:`{APPNAME_}`
          if it is :data:`Auto`.  If it is ``None``, no vars will be loaed.

    Returns:
        A dict with the loaded settings.
    """
    prefix = f"{appname.upper()}_" if isinstance(prefix, _Auto) else prefix
    if prefix is None:
        return {}

    env = os.environ
    values: Dict[str, Any] = {}
    for path, _field, _cls in fields:
        varname = f"{prefix}{path.upper().replace('.', '_')}"
        if varname in env:
            _set_path(values, path, env[varname])

    return values



"""
Test that all public functions are properly exposed.
"""
import pytest

import typed_settings as ts


@ts.settings
class Settings:
    u: str = ts.option()
    p: str = ts.secret()


def test_load_settings(tmp_path):
    """We can load settings with a class decorated with our decorator."""
    f = tmp_path.joinpath("cfg.toml")
    f.write_text('[test]\nu = "spam"\np = "eggs"\n')
    settings = ts.load_settings(Settings, "test", [f])
    assert settings == Settings("spam", "eggs")


def test_update_settings():
    s = Settings("spam", "eggs")
    u = ts.update_settings(s, "p", "bacon")
    assert u == Settings("spam", "bacon")


class TestAttrExtensions:
    """Tests for attrs extensions."""

    @pytest.fixture
    def inst(self):
        return Settings(u="spam", p="42")

    def test_secret_str(self, inst):
        assert str(inst) == "Settings(u='spam', p=***)"

    def test_secret_repr(self, inst):
        assert repr(inst) == "Settings(u='spam', p=***)"



"""
Typed settings
"""
from functools import partial

import attr

from ._click import click_options
from ._core import load_settings, update_settings


settings = attr.frozen
# settings = partial(attr.frozen, field_transformer=attr.auto_convert)
option = attr.field
secret = partial(attr.field, repr=lambda v: "***")


__all__ = [
    "load_settings",
    "update_settings",
    "click_options",
    "settings",
    "option",
    "secret",
]



import attr
import pytest

from typed_settings import _dict_utils as du


def mkattr(name: str, typ: type) -> attr.Attribute:
    """Creates an Attribute with *name* and *type*."""
    return attr.Attribute(  # type: ignore
        name, attr.NOTHING, None, True, None, None, True, False, type=typ
    )


class TestDeepFields:
    """Tests for _deep_fields()."""

    def test_deep_fields(self):
        @attr.dataclass
        class GrandChild:
            x: int

        @attr.dataclass
        class Child:
            x: float
            y: GrandChild

        @attr.dataclass
        class Parent:
            x: str
            y: Child
            z: str

        fields = du._deep_fields(Parent)
        assert fields == [
            ("x", mkattr("x", str), Parent),
            ("y.x", mkattr("x", float), Child),
            ("y.y.x", mkattr("x", int), GrandChild),
            ("z", mkattr("z", str), Parent),
        ]

    def test_unresolved_types(self):
        """Raise a NameError when types cannot be resolved."""

        @attr.dataclass
        class C:
            name: str
            x: "X"  # type: ignore  # noqa: F821

        with pytest.raises(NameError, match="name 'X' is not defined"):
            du._deep_fields(C)

    def test_direct_recursion(self):
        """
        We do not (and cannot easily) detect recursion.  A NameError is already
        raised when we try to resolve all types.  This is good enough.
        """

        @attr.dataclass
        class Node:
            name: str
            child: "Node"

        with pytest.raises(NameError, match="name 'Node' is not defined"):
            du._deep_fields(Node)

    def test_indirect_recursion(self):
        """
        We cannot (easily) detect indirect recursion but it is an error
        nonetheless.  This is not Dark!
        """

        @attr.dataclass
        class Child:
            name: str
            parent: "Parent"

        @attr.dataclass
        class Parent:
            name: str
            child: "Child"

        with pytest.raises(NameError, match="name 'Child' is not defined"):
            du._deep_fields(Parent)


@pytest.mark.parametrize(
    "path, expected",
    [
        ("a", 1),
        ("b.c", 2),
        ("b.d.e", 3),
        ("x", KeyError),
        ("b.x", KeyError),
    ],
)
def test_get_path(path, expected):
    """Tests for _get_path()."""
    dct = {
        "a": 1,
        "b": {
            "c": 2,
            "d": {
                "e": 3,
            },
        },
    }
    if isinstance(expected, int):
        assert du._get_path(dct, path) == expected
    else:
        pytest.raises(expected, du._get_path, dct, path)


def test_set_path():
    """We can set arbitrary paths, nested dicts will be created as needed."""
    dct = {}
    du._set_path(dct, "a", 0)
    du._set_path(dct, "a", 1)
    du._set_path(dct, "b.d.e", 3)
    du._set_path(dct, "b.c", 2)
    assert dct == {
        "a": 1,
        "b": {
            "c": 2,
            "d": {
                "e": 3,
            },
        },
    }


def test_dict_merge():
    """Dicts must be merged recursively.  Lists are just overridden."""
    d1 = {
        "1a": 3,
        "1b": {"2a": "spam", "2b": {"3a": "foo"}},
        "1c": [{"2a": 3.14}, {"2b": 34.3}],
        "1d": 4,
    }
    d2 = {
        "1b": {"2a": "eggs", "2b": {"3b": "bar"}},
        "1c": [{"2a": 23}, {"2b": 34.3}],
        "1d": 5,
    }
    du._merge_dicts(d1, d2)
    assert d1 == {
        "1a": 3,
        "1b": {"2a": "eggs", "2b": {"3a": "foo", "3b": "bar"}},
        "1c": [{"2a": 23}, {"2b": 34.3}],
        "1d": 5,
    }



import os
from itertools import product
from pathlib import Path
from typing import List

import pytest
from attr import field, frozen

from typed_settings import _core
from typed_settings._dict_utils import _deep_fields


@frozen
class Host:
    name: str
    port: int = field(converter=int)


@frozen(kw_only=True)
class Settings:
    url: str
    default: int = 3
    host: Host = field(
        converter=lambda d: Host(**d) if isinstance(d, dict) else d  # type: ignore  # noqa
    )


class TestAuto:
    """Tests for the AUTO sentinel."""

    def test_is_singleton(self):
        assert _core.AUTO is _core._Auto()

    def test_str(self):
        assert str(_core.AUTO) == "AUTO"


class TestLoadSettings:
    """Tests for load_settings()."""

    config = """[example]
        url = "https://example.com"
        [example.host]
        name = "example.com"
        port = 443
    """

    def test_load_settings(self, tmp_path, monkeypatch):
        """Test basic functionality."""
        monkeypatch.setenv("EXAMPLE_HOST_PORT", "42")

        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(self.config)

        settings = _core.load_settings(
            settings_cls=Settings,
            appname="example",
            config_files=[config_file],
        )
        assert settings == Settings(
            url="https://example.com",
            default=3,
            host=Host(
                name="example.com",
                port=42,
            ),
        )

    def test__load_settings(self, tmp_path, monkeypatch):
        """
        The _load_settings() can be easier reused.  It takes the fields lists
        and returns the settings as dict that can still be updated.
        """
        monkeypatch.setenv("EXAMPLE_HOST_PORT", "42")

        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(self.config)

        settings = _core._load_settings(
            fields=_deep_fields(Settings),
            appname="example",
            config_files=[config_file],
            config_file_section=_core.AUTO,
            config_files_var=_core.AUTO,
            env_prefix=_core.AUTO,
        )
        assert settings == {
            "url": "https://example.com",
            # "default": 3,  # This is from the cls!
            "host": {
                "name": "example.com",
                "port": "42",  # Value not yet converted
            },
        }


class TestUpdateSettings:
    """Tests for update_settings()."""

    settings = Settings(url="a", host=Host("h", 3))

    def test_update_top_level(self):
        """Top level attributes can be updated."""
        updated = _core.update_settings(self.settings, "default", 4)
        assert updated.default == 4
        assert updated.host == self.settings.host

    def test_update_nested_scalar(self):
        """Nested scalar attributes can be updated."""
        updated = _core.update_settings(self.settings, "host.name", "x")
        assert updated.host.name == "x"

    def test_update_nested_settings(self):
        """Nested scalar attributes can be updated."""
        updated = _core.update_settings(
            self.settings, "host", Host("spam", 23)
        )
        assert updated.host.name == "spam"
        assert updated.host.port == 23

    def test_copied(self):
        """
        Top level and nested settings classes are copied, even if unchanged.
        """
        updated = _core.update_settings(self.settings, "url", "x")
        assert updated is not self.settings
        assert updated.host is not self.settings.host

    @pytest.mark.parametrize("path", ["x", "host.x", "host.name.x"])
    def test_invalid_path(self, path):
        """
        Raise AttributeError if path points to non existing attribute.
        Improve default error message.
        """
        with pytest.raises(
            AttributeError,
            match=(f"'Settings' object has no setting '{path}'"),
        ):
            _core.update_settings(self.settings, path, "x")


class TestFromToml:
    """Tests for _from_toml()"""

    @pytest.fixture
    def fnames(self, tmp_path: Path) -> List[Path]:
        p0 = tmp_path.joinpath("0.toml")
        p1 = tmp_path.joinpath("1.toml")
        p2 = tmp_path.joinpath("2")
        p3 = tmp_path.joinpath("3")
        p0.touch()
        p2.touch()
        return [p0, p1, p2, p3]

    @pytest.mark.parametrize(
        "cfn, env, expected",
        [
            ([], None, []),
            ([0], None, [0]),
            ([1], None, []),
            ([2], None, [2]),
            ([3], None, []),
            ([], [0], [0]),
            ([0, 1], [2, 3], [0, 2]),
            ([2, 1, 0], [2], [2, 0, 2]),
        ],
    )
    def test_get_config_filenames(
        self, cfn, env, expected, fnames, monkeypatch
    ):
        """
        Config files names (cnf) can be specified explicitly or via an env var.
        It's no problem if a files does not exist (or is it?).
        """
        if env is not None:
            monkeypatch.setenv("CF", ":".join(str(fnames[i]) for i in env))
            env = "CF"

        paths = _core._get_config_filenames([fnames[i] for i in cfn], env)
        assert paths == [fnames[i] for i in expected]

    def test_load_toml(self, tmp_path):
        """We can load settings from toml."""

        @frozen
        class Sub:
            b: str

        @frozen
        class Settings:
            a: str
            sub: Sub

        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[example]
            a = "spam"
            [example.sub]
            b = "eggs"
        """
        )
        results = _core._load_toml(
            _deep_fields(Settings), config_file, "example"
        )
        assert results == {
            "a": "spam",
            "sub": {"b": "eggs"},
        }

    def test_load_from_nested(self, tmp_path):
        """
        We can load settings from a nested section (e.g., "tool.example").
        """

        @frozen
        class Sub:
            b: str

        @frozen
        class Settings:
            a: str
            sub: Sub

        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[tool.example]
            a = "spam"
            [tool.example.sub]
            b = "eggs"
        """
        )
        results = _core._load_toml(
            _deep_fields(Settings), config_file, "tool.example"
        )
        assert results == {
            "a": "spam",
            "sub": {"b": "eggs"},
        }

    def test_section_not_found(self, tmp_path):
        """
        An empty tick is returned when the config file does not contain the
        desired section.
        """

        @frozen
        class Settings:
            pass

        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[tool]
            a = "spam"
        """
        )
        assert _core._load_toml(Settings, config_file, "tool.example") == {}

    def test_load_convert_dashes(self, tmp_path):
        """
        Dashes in settings and section names are replaced with underscores.
        """

        @frozen
        class Sub:
            b_1: str

        @frozen
        class Settings:
            a_1: str
            a_2: str
            sub_section: Sub

        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[example]
            a-1 = "spam"
            a_2 = "eggs"
            [example.sub-section]
            b-1 = "bacon"
        """
        )
        results = _core._load_toml(
            _deep_fields(Settings), config_file, "example"
        )
        assert results == {
            "a_1": "spam",
            "a_2": "eggs",
            "sub_section": {"b_1": "bacon"},
        }

    def test_clean_settings(self):
        """
        Settings for which there is no attribute must be recursively removed.
        """
        settings = {
            "url": "abc",
            "host": {"port": 23, "eggs": 42},
            "spam": 23,
        }
        result = _core._clean_settings(_deep_fields(Settings), settings)
        assert result == {
            "url": "abc",
            "host": {"port": 23},
        }

    def test_clean_settings_unresolved_type(self):
        """
        Cleaning must also work if an options type is an unresolved string.
        """

        @frozen
        class Host:
            port: int = field(converter=int)

        @frozen(kw_only=True)
        class Settings:
            host: "Host" = field(
                converter=lambda d: Host(**d) if isinstance(d, dict) else d
            )

        settings = {"host": {"port": 23, "eggs": 42}}
        result = _core._clean_settings(_deep_fields(Settings), settings)
        assert result == {"host": {"port": 23}}

    def test_load_settings_explicit_config(self, tmp_path, monkeypatch):
        """
        The automatically derived config section name and settings files var
        name can be overriden.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[le-section]
            spam = "eggs"
        """
        )

        monkeypatch.setenv("LE_SETTINGS", str(config_file))

        @frozen
        class Settings:
            spam: str = ""

        settings = _core._from_toml(
            _deep_fields(Settings),
            appname="example",
            files=[],
            section="le-section",
            var_name="LE_SETTINGS",
        )
        assert settings == {"spam": "eggs"}

    @pytest.mark.parametrize(
        "is_mandatory, is_path, in_env, exists",
        product([True, False], repeat=4),
    )
    def test_mandatory_files(
        self,
        is_mandatory,
        is_path,
        in_env,
        exists,
        tmp_path,
        monkeypatch,
    ):
        """
        Paths with a "!" are mandatory and an error is raised if they don't
        exist.
        """
        p = tmp_path.joinpath("s.toml")
        if exists:
            p.touch()
        p = f"!{p}" if is_mandatory else str(p)
        if is_path:
            p = Path(p)
        files = []
        if in_env:
            monkeypatch.setenv("TEST_SETTINGS", str(p))
        else:
            files = [p]

        args = ([], "test", files, _core.AUTO, _core.AUTO)
        if is_mandatory and not exists:
            pytest.raises(FileNotFoundError, _core._from_toml, *args)
        else:
            _core._from_toml(*args)


class TestFromEnv:
    """Tests for _from_env()"""

    @pytest.mark.parametrize("prefix", ["T_", _core.AUTO])
    def test_from_env(self, prefix, monkeypatch):
        """Ignore env vars for which no settings attrib exis_core."""
        fields = _deep_fields(Settings)
        monkeypatch.setattr(
            os,
            "environ",
            {
                "T_URL": "foo",
                "T_HOST": "spam",  # Haha! Just a deceit!
                "T_HOST_PORT": "25",
            },
        )
        settings = _core._from_env(fields, "t", prefix)
        assert settings == {
            "url": "foo",
            "host": {
                "port": "25",
            },
        }

    def test_no_env_prefix(self, monkeypatch):
        """
        The prefix for env vars can be disabled w/o disabling loading env. vars
        themselves.
        """
        monkeypatch.setenv("CONFIG_VAL", "42")

        @frozen
        class Settings:
            config_val: str

        settings = _core._from_env(_deep_fields(Settings), "example", "")
        assert settings == {"config_val": "42"}

    def test_disable_environ(self):
        """Setting env_prefix=None diables loading env vars."""

        @frozen
        class Settings:
            x: str = "spam"

        settings = _core._from_env(_deep_fields(Settings), "example", None)
        assert settings == {}



from setuptools import find_packages, setup


DEPS = ["attrs", "toml"]
DEPS_CLICK = ["click"]
DEPS_TEST = DEPS_CLICK + ["pytest>=6", "pytest-cov", "coverage[toml]>=5.0.2"]
DEPS_LINT = DEPS_CLICK + [
    "flake8",
    "flake8-bandit",
    "flake8-black",
    "flake8-bugbear",
    "flake8-isort",
    "mypy",
]
DEPS_DEV = DEPS_TEST + DEPS_LINT + ["nox", "safety"]


if __name__ == "__main__":
    setup(
        name="typed-settings",
        version="0.4",
        description="Typed settings based on attrs classes",
        license="MIT",
        url="https://gitlab.com/sscherfke/typed-settings",
        project_urls={
            "Documentation": "https://gitlab.com/sscherfke/typed-settings",
            "Bug Tracker": "https://gitlab.com/sscherfke/typed-settings/-/issues",
            "Source Code": "https://gitlab.com/sscherfke/typed-settings",
        },
        author="Stefan Scherfke",
        author_email="stefan@sofa-rockers.org",
        maintainer="Stefan Scherfke",
        maintainer_email="stefan@sofa-rockers.org",
        keywords=["settings", "types", "configuration", "options"],
        long_description=open("README.md").read(),
        long_description_content_type="text/markdown",
        packages=find_packages(where="src"),
        package_dir={"": "src"},
        include_package_data=True,
        zip_safe=False,
        python_requires=">=3.7",
        install_requires=DEPS,
        extras_require={
            "click": DEPS_CLICK,
            "test": DEPS_TEST,
            "lint": DEPS_LINT,
            "dev": DEPS_DEV,
        },
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "Natural Language :: English",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: Implementation :: CPython",
            "Programming Language :: Python :: Implementation :: PyPy",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    )
